#include "foofighter.h"

/* These functions make sure that all the various game components are
 * updated correctly. */

/* The status message display record. */

static struct ave_imgdlist_rec_string statusmsg;
static int statusmsgtime;

static void animate_statusmsg(void);

/* The string display routines need a ruler to tell it where the tab stops
 * are. We don't use this, so the ruler is empty.
 */

static int ruler[] = {0};

/* Initialise everything. */

void init_animator(void)
{
	/* Initialise the various components of the game. */

	init_repositories();
	init_stars();

	/* Initialise the status message. Defaults to empty. */

	statusmsg.rec.type = IMGDLIST_TYPE_STRING;
	statusmsg.rec.x = 0;
	statusmsg.rec.y = 0;
	statusmsg.rec.width = 0;
	statusmsg.rec.height = 0;
	statusmsg.flags = TRF_INK;
	statusmsg.ink = 0x0000000;
	statusmsg.paper = 0xFF000000;
	statusmsg.pixrule = PIX_RULE_ALPHA;
	statusmsg.pString = "";
	statusmsg.pFont = font;
	statusmsg.charcnt = 0;
	statusmsg.pRuler = ruler;
	elate_reftool("ave/util/str/r_utf8", statusmsg.pDecoder);

	/* Add the status message to the display list. */

	ave_imgdlist_addrechead(explosiondisplaylist,
			(ave_imgdlist_rec_t*) &statusmsg);
}

/* This is called every frame. Animate all objects on the list one frame. */

void animator_tick(void)
{
	/* Animate everything. This must be the last thing that happens in the
	 * event loop; otherwise, you end up displaying half-initialised
	 * objects. */

	animate_stars();
	animate_objects();
	animate_statusmsg();
}

/* Initialise the animation repositories. Each repository is a long strip
 * containing the frames of the animation. Unfortunately, each one contains
 * a different number of frames. */

typedef struct {
	char* filename;		// Name of file containing repository
	int alpha;		// Use alpha blending?
	int frames;		// Number of frames
	int width;		// Width of a frame
	int height;		// Height of a frame
	ave_pix_t* avo;		// AVO pointer of repository
} repositoryentry_t;

static repositoryentry_t repositorydata[] = {
	{DIRPREFIX "ship.png",		0, 11, 0, 0,	NULL}, // SPRITE_SHIP
	{DIRPREFIX "explosion.png",	1, 29, 0, 0,	NULL}, // SPRITE_EXPLOSION
	{DIRPREFIX "spinner.png",	0, 15, 0, 0,	NULL}, // SPRITE_SPINNER
	{DIRPREFIX "bullet.png",	0, 1, 0, 0,	NULL}, // SPRITE_BULLET
	{DIRPREFIX "redflash.png",	0, 6, 0, 0,	NULL}, // SPRITE_REDFLASH
	{DIRPREFIX "greenflash.png",	0, 6, 0, 0,	NULL}, // SPRITE_GREENFLASH
	{DIRPREFIX "blueflash.png",	0, 6, 0, 0,	NULL}, // SPRITE_BLUEFLASH
	{DIRPREFIX "stoneball.png",	0, 11, 0, 0,	NULL}, // SPRITE_STONEBALL
	{DIRPREFIX "spikedwheel.png",	0, 15, 0, 0,	NULL}, // SPRITE_SPIKEDWHEEL
	{DIRPREFIX "ringball.png",	0, 11, 0, 0,	NULL}, // SPRITE_RINGBALL
	{DIRPREFIX "d12.png",		0, 19, 0, 0,	NULL}, // SPRITE_D12
	{DIRPREFIX "cube.png",		0, 16, 0, 0,	NULL}, // SPRITE_CUBE
	{DIRPREFIX "coin.png",		0, 20, 0, 0, 	NULL}, // SPRITE_COIN
};
#define REPOSITORY_LENGTH (sizeof(repositorydata)/sizeof(*repositorydata))

void init_repositories(void)
{
	int i;

	for (i=0; i<REPOSITORY_LENGTH; i++)
	{
		ave_pix_t* pix;
		int depth;

		/* What pixelmap do we want to load the images in? Good
		 * question. For optimum speed, we want to use the same type
		 * as the screen itself, to avoid format conversions when
		 * blitting. */

		depth = ave_dev_getscreen(ave).type;
		
		/* However, if any image is marked as being alpha blended, it
		 * needs to go in an alpha-blended pixelmap. There are two: 12
		 * bit and 32 bit. We're going to use 12 bit. It's a bit
		 * slower than 32 bit but uses half the memory. */

		if (repositorydata[i].alpha)
			depth = PIX_TYPE_12;

		/* Load the image off disk. */

		pix = ave_cnv_share(
				repositorydata[i].filename,	// Filename
				depth,				// Depth
				FCNV_HWACCEL			// In video RAM
			);
		if (!pix)
			fatalerror("Failed to load %s", repositorydata[i].filename);

		/* Now we need to tell the image that it uses itself as its
		 * own mask, with colour 0 signifying transparent. This
		 * applies to both normal and alpha-ed images. For alpha-ed
		 * images, it means we only have to do the expensive alpha
		 * blending bit for pixels that are not totally transparent. */

		ave_pix_setmask(pix, pix, 0, 0);
		ave_pix_settrans(pix, 0x000000);

		/* If the image is alpha-ed, turn on alpha blending. */

		if (repositorydata[i].alpha)
			ave_avo_setflags((ave_avo_t*) pix, FAVO_ALPHA, FAVO_ALPHA);

		/* Work out the width and height of a frame. */

		{
			ave_avo_size_t r = ave_avo_getsize((ave_avo_t*) pix);
			repositorydata[i].width = r.width / repositorydata[i].frames;
			repositorydata[i].height = r.height;
		}

		/* And set it. */

		repositorydata[i].avo = pix;
	}
	printf("\n");
}

/* Free the repositories once we're done with them. While the repositories are
 * AVOs, they're not attached to our application object in any way, so they
 * won't be freed when the application object is. */

void free_repositories(void)
{
	int i;

	for (i=0; i<sizeof(repositorydata)/sizeof(*repositorydata); i++)
		if (repositorydata[i].avo)
			ave_obj_deref((ave_obj_t*) repositorydata[i].avo);
}

/* Initialise a dlist record given the sprite. */

void init_sprite(object_t* object)
{
	repositoryentry_t* r = &repositorydata[object->sprite];

	object->drec.rec.type = IMGDLIST_TYPE_BLIT;
	object->drec.rec.width = r->width;
	object->drec.rec.height = r->height;
	object->drec.pSource = r->avo;

	ave_imgdlist_addrechead(object->layer,
			(ave_imgdlist_rec_t*) &object->drec);
}

/* Update a sprite's dlist record. */

void update_sprite(object_t* object)
{
	repositoryentry_t* r = &repositorydata[object->sprite];

	/* Move the object according to its velocity. */

	object->rx += object->xv;
	object->ry += object->yv;

	/* Sanitise the frame number. */

	if (object->frame >= r->frames)
		object->frame -= r->frames;
	else if (object->frame < 0)
		object->frame += r->frames;

	/* And set the dlist record according to the position. */

	object->drec.sx = object->frame * r->width;
	object->drec.sy = 0;
	object->drec.rec.x = (object->rx >> 16) - r->width/2;
	object->drec.rec.y = (object->ry >> 16) - r->height/2;
}

/* Set the status message. */

void set_statusmsg(char* msg)
{
	/* First, work out the size of the message. */

	ave_font_dimensions_t r = ave_font_strsize(font, msg, NULL);

	/* Set the display record up with the message centred on the screen.
	 */

	statusmsg.rec.x = WIN_WIDTH/2 - r.width/2;
	statusmsg.rec.y = WIN_HEIGHT/2 - r.height/2;
	statusmsg.rec.width = r.width;
	statusmsg.rec.height = r.height;
	statusmsg.pString = msg;
	statusmsg.charcnt = strlen(msg);

	/* Set the counter. */

	statusmsgtime = 0;
}

/* Animate the status message. */

void animate_statusmsg(void)
{
	if (statusmsgtime == -1)
	{
		/* No message. */
		return;
	}
	
	statusmsgtime++;

	/* Set the colour. Uses nifty interpolation effects to make the
	 * message look good. */

	#define EPOCH (FRAMERATE/2)
	if (statusmsgtime < EPOCH)
		statusmsg.ink = interpolateRGB(0x00000000, 0x00FF0000,
				0, EPOCH, statusmsgtime);
	else if (statusmsgtime < EPOCH*2)
		statusmsg.ink = interpolateRGB(0x00FF0000, 0x00FFFF00,
				EPOCH, EPOCH*2, statusmsgtime);
	else if (statusmsgtime < EPOCH*3)
		statusmsg.ink = interpolateRGB(0x00FFFF00, 0xFF000000,
				EPOCH*2, EPOCH*3, statusmsgtime);
	else
	{
		/* Turn off the message. */

		statusmsg.rec.x = 0;
		statusmsg.rec.y = 0;
		statusmsg.rec.width = 0;
		statusmsg.rec.height = 0;
		statusmsgtime = -1;
	}
}

