#include "foofighter.h"

/* This code contains all the collision detection stuff.
 *
 * We support two different collision detection schemes, selectable at compile
 * time. RECTANGULAR_REGIONS checks to see if the two objects' bounding
 * rectangles overlap. CIRCULAR_REGIONS assumes each object is a circle with
 * the diameter of the width of the bounding rectangle and sees if *they*
 * overlap. You must define one or else no collision detection will be done!
 */

//#define RECTANGULAR_REGIONS
#define CIRCULAR_REGIONS

/* Check to see if the two objects overlap. */

static int hits(object_t* obj1, object_t* obj2)
{
#ifdef RECTANGULAR_REGIONS
	/* Calculates the minimum possible distance for collision in both X
	 * and Y and checks them against the real distances. */

	int o1w = obj1->drec.rec.width << 16;
	int o1h = obj1->drec.rec.height << 16;
	int o2w = obj2->drec.rec.width << 16;
	int o2h = obj2->drec.rec.height << 16;

	/* The minimum width and height between the objects must be the
	 * sum of the halves of the two. */

	int w = (o1w + o2w)/2;
	int h = (o1h + o2h)/2;

	/* Calculate the actual minimum distance. */

	int dw = abs(obj1->rx - obj2->rx);
	int dh = abs(obj1->ry - obj2->ry);

	/* There's only a collision if the distance is less than or equal to
	 * the minimum. */

	return (dw <= w) && (dh <= h);
#endif
#ifdef CIRCULAR_REGIONS
	/* Calculates the minimum possible distance for collision in a
	 * straight line and checks it against the real distance.
	 *
	 * However, this algorithm is made more complicated by the fact that
	 * we need to multiply two 16.16 numbers together. This don't work! At
	 * least not if you are using ints! As 16.16 * 16.16 = 32.32, you end
	 * up dropping all the integer part and you're just left with the
	 * fractional part. As we don't need the fractional part in this case,
	 * we do all our maths with 32 bit integers. */

	/* Determines the radius of the two circles. This is the width divided
	 * by two. */

	int r1 = obj1->drec.rec.width/2;
	int r2 = obj2->drec.rec.width/2;

	/* Minimum distance, squared, is the sum of the two radii. */

	int d2 = (r1+r2) * (r1+r2);

	/* Calculate the real distance, squared, using pythagoras. */

	int o1x = obj1->rx >> 16;
	int o1y = obj1->ry >> 16;
	int o2x = obj2->rx >> 16;
	int o2y = obj2->ry >> 16;
	int rd2 = (o1x - o2x)*(o1x - o2x) + (o1y - o2y)*(o1y - o2y);

	/* There's only a collision if the distance is less than or equal to
	 * the minimum. */

	return (rd2 <= d2);
#endif
}

/* Check the specified object to make sure it isn't being hit by any ordnance.
 */

static int collision_detect_against_ordnance(object_t* obj)
{
	ELATE_LIST_NODE* n;
	ELATE_LIST_NODE* nextn;

	/* Iterate through all objects looking for ordnance.. */

	n = elate_headend(&objectlisthead);
	nextn = elate_succ(n);
	while (elate_succnode(nextn, n))
	{
		object_t* o = (object_t*) n;

		if (obj == o)
			continue;

		if (!o->ordnance)
			continue;

		if (!hits(o, obj))
			continue;

		/* o does hit obj. Damage obj. */
		
		obj->damageable--;

		/* Has obj been destroyed? */

		if (obj->damageable == 0)
		{
			/* Yes; remove it and create an explosion. */

			play_sample(SOUND_EXPLOSION);
			create_explosion(SPRITE_EXPLOSION, obj->rx, obj->ry);
			adjust_bank(obj->bonus);
			destroy_object(obj);
		}
		else
		{
			/* No; nudge obj and create a flash. */

			if (o->rx < obj->rx)
				obj->xv += 1<<16;
			else
				obj->xv -= 1<<16;
			obj->yv -= 1<<16;

			{
				static int randomflash[] = {
					SPRITE_REDFLASH,
					SPRITE_GREENFLASH,
					SPRITE_BLUEFLASH
				};

				play_sample(SOUND_FLASH);
				create_explosion(randomentry(randomflash),
						obj->rx, obj->ry);
			}
		}

		/* Remember to destroy the bullet. */

		destroy_object(o);

		return 1;
	}

	return 0;
}

/* Check the specified object to make sure it's not hitting the player. */

static int collision_detect_against_player(object_t* obj)
{
	if (!player)
		return 0;

	if (!hits(obj, player))
		return 0;

	/* The player dies! */

	play_sample(SOUND_EXPLOSION);
	play_sample(SOUND_FLASH);
	create_explosion(SPRITE_EXPLOSION, player->rx, player->ry);
	create_explosion(SPRITE_EXPLOSION, obj->rx, obj->ry);
	destroy_object(obj);
	destroy_object(player);
	player = NULL;
	
	return 1;
}

/* Collision detect all the objects. */

void collision_detect(void)
{
	ELATE_LIST_NODE* n;
	ELATE_LIST_NODE* nextn;

	/* Iterate through all objects. */

restart:
	n = elate_headend(&objectlisthead);
	nextn = elate_succ(n);
	while (elate_succnode(nextn, n))
	{
		object_t* o = (object_t*) n;

		/* Now we collision detect each object depending what it's
		 * subject to. If a collision detect failed, we need to
		 * restart the loop because the list will have changed (one or
		 * two objects removed, an explosion object added). */

		/* Damageable by ordnance? */
		
		if (o->damageable && collision_detect_against_ordnance(o))
			goto restart;

		/* Can it hurt the player? */

		if (o->deadly && collision_detect_against_player(o))
			goto restart;
	}
}

