#ifndef FOOFIGHTER_H

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <elate/elate.h>
#include <elate/kn.h>
#include <elate/ave/ave.h>

/* Configuration */

//#define SOUND				// Use AOM sound effects?
//#define MUSIC				// Use Koan music?

#define DIRPREFIX "/cowlark/foofighter/" // Directory containing game
#define WIN_WIDTH 640			// Width of window
#define WIN_HEIGHT 480			// Height of window
#define FRAMERATE 40			// Number of frames per second

#define ACCELERATION 0x30000		// Player acceleration
#define DAMPING 20			// Player velocity damping (% per frame)
#define BULLETSPEED (15<<16)		// How fast do bullets go?

#define NUMBER_OF_STARS 100		// Number of background stars

#define MISSED_ASTEROID_PENALTY	100	// Amount you're fined if you miss an asteroid
#define SHOT_PENALTY 1			// Amount you lose for every shot
#define DESTROYED_ASTEROID_BONUS 10	// Amount you make when you destroy an asteroid

/* Global variables */

extern ave_dev_t* ave;			// The AVE itself
extern ave_app_t* app;			// Our application object
extern ave_toolkit_t* toolkit;		// Our toolkit
extern ave_font_t* font;		// Message font

extern ave_win_t* mainwindow;		// The application window
extern ave_avo_t* content;		// The window's content area

extern ave_img_t* stardisplaylist;	// The star display list
extern ave_img_t* scenerydisplaylist;	// Scenery
extern ave_img_t* weapondisplaylist;	// Ordinance
extern ave_img_t* shipdisplaylist;	// Ships (including the player)
extern ave_img_t* explosiondisplaylist;	// Explosions

/* Keyboard flags */

extern int key_left;			// Left pressed?
extern int key_right;			// Right pressed?
extern int key_up;			// Up pressed?
extern int key_down;			// Down pressed?
extern int key_fire;			// Fire pressed?

/* Game state */

extern int money;			// Amount of money in the bank
extern int distance;			// Number of metres travelled

/* User events. */

enum {
	UEV_TICK = EV_USER
};

/* Sprite numbers. Must match the order of the repository data store in
 * animator.c. */

enum {
	SPRITE_SHIP,
	SPRITE_EXPLOSION,
	SPRITE_SPINNER,
	SPRITE_BULLET,
	SPRITE_REDFLASH,
	SPRITE_GREENFLASH,
	SPRITE_BLUEFLASH,
	SPRITE_STONEBALL,
	SPRITE_SPIKEDWHEEL,
	SPRITE_RINGBALL,
	SPRITE_D12,
	SPRITE_CUBE,
	SPRITE_COIN
};

/* Sound sample numbers. Must match the order of the repository data store in
 * audio.c. */

enum {
	SOUND_EXPLOSION,
	SOUND_SHOT,
	SOUND_FLASH
};

/* Describes an object. */

typedef struct object_t {
	ELATE_LIST_NODE ln;		// Linked list metadata

	/* Generic data */

	int sprite;			// Which sprite to use?
	int frame;			// Which frame to use?
	int rx, ry;			// 16.16 position of object
	int xv, yv;			// 16.16 velocity of object
	int deadly;			// Does this object hurt the player?
	int damageable;			// Can this object be hurt by ordinance?
	int ordnance;			// Is this some ordnance?
	int bonus;			// Amount you make when you destroy it
	int penalty;			// Amount you lose if you miss it
	ave_img_t* layer;		// The layer the object lives in
	ave_imgdlist_rec_blit_t drec;	// This object's dlist record
	
	/* Object-specific data */

	union {
		struct {
			int firing;	// Fire key pressed?
		} player;
	} u;

	/* Methods */

	void (*tick)(struct object_t* object);		// Process a frame
	void (*deinit)(struct object_t* object);	// Clean up an object
} object_t;

/* Returns a random entry from the given array. */

#define randomentry(a) \
	(a)[randomrange(sizeof(a)/sizeof(*a))]

/* intent doesn't have these */

#define max(a, b) \
	((a) < (b) ? (b) : (a))

#define min(a, b) \
	((a) > (b) ? (b) : (a))

/* Game mechanics. */

extern void aveevent(long wakeuptime);
extern void new_player(void);
extern void player_died(void);
extern void player_insolvent(void);
extern void start_game(void);
extern void tick(void);

/* The object list. */

extern ELATE_LIST_HEAD objectlisthead;
extern object_t* player;

/* Function prototypes --- Utilities */

extern void fatalerror(char* msg, ...);
extern int randomrange(int size);
extern int interpolate(int A, int B, int a, int b, int z);
extern int interpolateRGB(int A, int B, int a, int b, int z);

/* The animator */

extern void init_animator(void);
extern void animator_tick(void);

extern void init_repositories(void);
extern void free_repositories(void);

extern void init_sprite(object_t* object);
extern void update_sprite(object_t* object);
extern void destroy_sprite(object_t* object);

extern void set_statusmsg(char* msg);

/* Audio */

extern void init_music(void);
extern void shutdown_music(void);
extern void start_music(void);
extern void stop_music(void);

extern void init_audio(void);
extern void shutdown_audio(void);
extern void play_sample(int sample);

/* Stars */

extern void init_stars(void);
extern void animate_stars(void);

/* Generic objects */

extern void init_objects(void);
extern void animate_objects(void);
extern object_t* alloc_object(void);
extern void destroy_object(object_t* object);

/* Collision detection */

extern void collision_detect(void);

/* The player */

extern object_t* create_player(int x, int y);
extern void tick_player(object_t* object);

/* Drifters */

extern object_t* create_drifter(int sprite, ave_img_t* layer, int x, int y,
		int xv, int yv);
extern void tick_drifter(object_t* object);

/* Bullets */

extern object_t* create_bullet(int type, int x, int y, int xv, int yv);
extern void tick_bullet(object_t* object);

/* Explosions */

extern object_t* create_explosion(int sprite, int x, int y);
extern void tick_explosion(object_t* object);

/* The bank */

extern void init_bank(void);
extern void reset_bank(void);
extern void adjust_bank(int amount);
extern void adjust_progress(int delta);

/* Main program */

extern void shutdown(void);

#endif

