#include "foofighter.h"
#include "dev/keyboard/keyboard.h"

/* Lots of pointers to important AVE objects are defined here. Some of these
 * must be manually freed when the application exits; some are freed
 * automatically. The ones which are defined to be NULL are the ones we need
 * to free ourselves. See shutdown() for the code that actually does the
 * freeing. */

/* Global variables are defined here. */

ave_dev_t* ave = NULL;			// The AVE itself
ave_app_t* app = NULL;			// Our application object
ave_toolkit_t* toolkit = NULL;		// Our toolkit
ave_font_t* font = NULL;		// On-screen message font
ave_message_t* wakeupmsg = NULL;	// The timer message

/* On-screen gadgets. */

ave_win_t* mainwindow;
ave_avo_t* content;
ave_img_t* stardisplaylist;
ave_img_t* scenerydisplaylist;
ave_img_t* weapondisplaylist;
ave_img_t* shipdisplaylist;
ave_img_t* explosiondisplaylist;

/* Keyboard flags. */

int key_left;
int key_right;
int key_up;
int key_down;
int key_fire;

/* Free all resources we've got from the AVE. */

void shutdown(void)
{
	/* Shutdown the audio layer. */

	shutdown_music();
	shutdown_audio();

	/* Free the wakeup event. */

	if (wakeupmsg)
		ave_dev_freeevent(ave, wakeupmsg);

	/* Free animation repositories. */

	free_repositories();

	/* Free the font. */

	if (font)
		ave_obj_deref((ave_obj_t*) font);

	if (toolkit)
	{
		/* Close all gadgets created by this application. */
		ave_app_closeall(app);

		/* Shut down and close the toolkit. */
		ave_dev_closetoolkit(ave, toolkit);
	}

	/* Shutdown and destroy the application. The ave itself doesn't need
	 * closing. */

	if (ave && app)
		ave_dev_close(ave, app);
}

/* Open the AVE and create the screen-visible gadgets. */

void init_screen(void)
{
	/* Find the AVE and create the application object. */

	{
		ave_dev_data_t r = ave_dev_lookup("device/ave/foofighter");
		if (iselateerrno(r.ave))
			fatalerror("failed to open AVE");

		ave = r.ave;
		app = ave_dev_open(ave, r.residual, 0, 0);
		if (iselateerrno(app))
			fatalerror("failed to create application object");
	}

	/* Open the toolkit. */

	toolkit = ave_dev_opentoolkit(ave, app);
	if (toolkit == NULL)
		fatalerror("failed to open toolkit");

	/* Create the on-screen message font. */

	font = ave_font_open("Gill Sans", 30, 0);
	if (font == NULL)
		fatalerror("failed to open font");
	
	/* Create the main window. */

	mainwindow = ave_toolkit_createdialog(toolkit,
			"Scroller Example",		/* title */
			NULL,				/* title pixelmap */
			WIN_WIDTH, WIN_HEIGHT,		/* width, height */
			FDI_FILLED | FDI_INNER | FDI_CLOSE | FDI_BORDER |
			 FDI_TITLE | FDI_CONTENT | FDI_DRAG | FDI_DEPTH /* furniture */
		);
	if (mainwindow == NULL)
		fatalerror("failed to create main window");

	/* We need to wire up the window's close button so that it sends us an
	 * EV_QUIT, as a basic way of making our app exit. */

	ave_avo_addlink((ave_avo_t*) mainwindow, (ave_evt_t*) app,
			CH_DIALOG_ACTION, EV_QUIT);

	/* Next, we need to tell the AVE that this window is going to be
	 * receiving keyboard events. We do this by giving it the keyboard
	 * token. */

	ave_avo_settoken((ave_avo_t*) mainwindow, BFAVO_KEYTOKEN);

	/* Get the content area of the window. This is the thing that you put
	 * the window's children in. */

	content = ave_dialog_getgadgets(mainwindow).content;

	/* Now we need to create the stack of AVOs that the user will look at.
	 * The bottom layer is the background, which should be dark. To make
	 * things look a little better, we'll use a vgrad widget that'll give
	 * us a nice gradient. We add everything with an index of 0. This puts
	 * each AVO at the top of the stack, so we end up defining everything
	 * from the bottom up. */

	{
		ave_img_t* img = ave_img_openvgrad(WIN_WIDTH, WIN_HEIGHT,
				0x000000,			/* top */
				0x000040			/* bottom */
			);
		ave_avo_add(content, (ave_avo_t*) img, 0);
	}

	/* Create the display lists and add it to the content. */

	stardisplaylist = ave_img_opendlist(WIN_WIDTH, WIN_HEIGHT);
	ave_avo_add(content, (ave_avo_t*) stardisplaylist, 0);

	scenerydisplaylist = ave_img_opendlist(WIN_WIDTH, WIN_HEIGHT);
	ave_avo_add(content, (ave_avo_t*) scenerydisplaylist, 0);

	weapondisplaylist = ave_img_opendlist(WIN_WIDTH, WIN_HEIGHT);
	ave_avo_add(content, (ave_avo_t*) weapondisplaylist, 0);

	shipdisplaylist = ave_img_opendlist(WIN_WIDTH, WIN_HEIGHT);
	ave_avo_add(content, (ave_avo_t*) shipdisplaylist, 0);

	explosiondisplaylist = ave_img_opendlist(WIN_WIDTH, WIN_HEIGHT);
	ave_avo_add(content, (ave_avo_t*) explosiondisplaylist, 0);

	/* ...add the window to the application object... */

	ave_app_addavo(app,
			NULL,				/* add to screen */
			(ave_avo_t*) mainwindow,	/* our window */
			-1				/* next index */
		);

	/* ...and update it. Otherwise it won't be drawn and we'll get our
	 * content area floating in space until someone moves another window
	 * over it and it gets exposed. */

	ave_avo_update((ave_avo_t*) mainwindow);

	/* Create the timer message. We send a message to ourselves every
	 * FRAMERATEth of a second. Rather than allocate a new one and free it
	 * every time, we can reuse the same message over and over. */

	wakeupmsg = ave_dev_allocevent(ave, 0);
}

/* The main program. */

int main(int argc, char* argv[])
{
	/* Open the AVE and initialise all subsystems. */

	init_screen();
	init_music();
	init_audio();
	init_animator();
	init_objects();
	init_bank();

	/* All keys are released. */

	key_left = 0;
	key_right = 0;
	key_up = 0;
	key_down = 0;

	/* Start a new game. */

	start_music();
	start_game();

	/* Never reached */

	exit(-1);
}

/* Process an AVE event.
 *
 * This routine is inside out compared to most AVE applications. Usually, AVE
 * apps go like this:
 *
 * do {
 * 	getevent();
 * 	doevent();
 * } while (not quitting);
 *
 * However, it makes our game logic a lot easier if we do:
 *
 * do {
 * 	gamelogic();
 * 	avestuff();
 * } while (not dead);
 *
 * So. This routine is called with an argument telling it when it needs to
 * return. It goes away and handles AVE events until the specified timeout
 * occurs, and then returns. This allows it to do nifty things like recycle
 * the timeout event so we don't have to keep allocating one, and makes the
 * game logic so much simpler you wouldn't believe it. (With a conventional
 * structure, I'd have to use a state machine. Ugh.)
 */

void aveevent(long wakeuptime)
{
	/* We want to queue a timed event using the wakeup time we've just
	 * been given. */

	wakeupmsg->EVD_TIMESTAMP = wakeuptime;
	wakeupmsg->EVD_TYPE = UEV_TICK;

	/* Now queue it. */

	ave_dev_addtimedevent(ave, (ave_evt_t*) app, wakeupmsg);

	/* Now we have the event loop. This fetches events from the AVE and
	 * processes them one by one. When we receive a tick event, we return.
	 */

	for (;;)
	{
		/* Wait for an event. We use a timeout of -1, which means
		 * block forever or until the next event (usually, one of our
		 * timer events). */

		ave_event_t event = ave_app_getevent(app, -1);

		/* All events are targeted at something. If not, we timed out.
		 */

		if (event.target == NULL)
			continue;

		/* Process the event. */

		if (event.type == UEV_TICK)
		{
			/* It's a tick event. Return immediately. */
			return;
		}
		else if ((event.type == EV_KEYDOWN) || (event.type == EV_KEYUP))
		{
			/* Set the keyboard flags according to the key that's
			 * been pressed or released. */

			int updown = (event.type == EV_KEYDOWN);

			switch (event.msg->EVD_KEYCOOKED)
			{
				case ELATE_VK_UP:
					key_up = updown;
					break;

				case ELATE_VK_DOWN:
					key_down = updown;
					break;

				case ELATE_VK_LEFT:
					key_left = updown;
					break;

				case ELATE_VK_RIGHT:
					key_right = updown;
					break;

				case ELATE_VK_RETURN:
				case ' ':
					key_fire = updown;
					break;
			}
		}
		else if (event.type < EV_USER)
			ave_evt_event(event.target, event.msg, event.type);

		/* Have we been told to exit? If so, bail out instantly (boss
		 * mode). */

		if (event.type == EV_QUIT)
		{
			
			/* Tricky bit: preventing the wakeup event from being
			 * cleaned up. You see, the event is still queued. So
			 * we can't just free it or it'll crash horribly.
			 * Luckily, the AVE is quite intelligent. If it gets a
			 * message targeted at a bogus application, it'll
			 * discard it and free it automatically. So all we
			 * need to do here is prevent *us* from freeing the
			 * message. We only have to do this when exiting while
			 * a wakeup event is still queued. In all other cases,
			 * it needs to be properly cleaned up. */

			wakeupmsg = NULL;
			shutdown();
			exit(0);
		}
		
		/* Free the event message. This was allocated automatically by
		 * the thing that triggered the event. */

		ave_dev_freeevent(ave, event.msg);
	
		/* And continue. */
	}
}

