#include "foofighter.h"

/* This code handles the background stars.
 *
 * This is the only place in the entire game where we draw stuff on the
 * screen. This is because the stars, being single pixels, are rather overkill
 * for a display list. The first version of this code *did* implement them in
 * a display list with one record per star, but that was *so* ugly.
 *
 * So instead we use a display list with a single record in it. (The overhead
 * for a single record is negligable.) When the record is drawn, a callback
 * gets called and we plot all the stars.
 */

/* The single record itself. */

struct ave_imgdlist_rec_user dlistrecord;

/* Information about each star. */

struct star_record {
	int rx;			/* X position, 16.16 */
	int ry;			/* Y position, 16.16 */
	int yv;			/* Stellar velocity, 16.16 */
	int colour;		/* Colour of this star. */
} star[NUMBER_OF_STARS];

/* Plot the stars. This is called by the AVE. */

static void plot_stars(ave_avo_t* avo, ave_ctx_t* ctx, ave_imgdlist_rec_t* dlistrec)
{
	int i;

	/* This is actually not nearly as efficient as it looks.
	 * ave_ctx_plot() does an ncall, which is quite expensive. In VP, I'd
	 * use the xmethod infrastructure to cache the pointer to the tool
	 * that actually did the plot and then gos it. However, there's no way
	 * of doing this in C (yet) so we have to do things the slow way. */

	for (i=0; i<NUMBER_OF_STARS; i++)
		ave_ctx_plot(ctx, star[i].rx >> 16, star[i].ry >> 16, star[i].colour);
}

/* Initialise the stars. */

void init_stars(void)
{
	int i;
	struct star_record* s;

	/* The list of colours that a star can be. */

	static int colours[] = {
		0x000000,	/* Speed 0, never used */
		0x707070,	/* Speed 1 */
		0x909090,	/* Speed 2 */
		0xB0B0B0,	/* Speed 3 */
		0xD0D0D0,	/* Speed 4 */
		0xFFFFFF	/* Speed 5, fastest */
	};

	/* Firstly, initialise the dlist record and add it. */

	dlistrecord.rec.type = IMGDLIST_TYPE_USER;
	dlistrecord.rec.x = 0;
	dlistrecord.rec.y = 0;
	dlistrecord.rec.width = WIN_WIDTH;
	dlistrecord.rec.height = WIN_HEIGHT;
	dlistrecord.callback = plot_stars;
	ave_imgdlist_addrechead(stardisplaylist, (ave_imgdlist_rec_t*) &dlistrecord);

	/* Now initialise the stars themselves. */

	for (i=0; i<NUMBER_OF_STARS; i++)
	{
		/* Work out the basic attributes for each star. */

		s = &star[i];
		s->rx = rand() % (WIN_WIDTH<<16);
		s->ry = rand() % (WIN_HEIGHT<<16);

		/* Work out a velocity and set the colour appropriately. */

		s->yv = (rand() % (4<<16)) + (1<<16);
		s->colour = colours[s->yv >> 16];
	}
}

/* Moves the stars on one tick. */

void animate_stars(void)
{
	int i;
	struct star_record* s;

	for (i=0; i<NUMBER_OF_STARS; i++)
	{
		s = &star[i];

		/* Move the star according to the velocity. If it falls off
		 * the bottom of the screen, put it back again at the top. */
	
		s->ry += s->yv;
		if (s->ry > (WIN_HEIGHT<<16))
			s->ry -= (WIN_HEIGHT<<16);
	}
}
